#-------------------------------------------------------------------------------
# Name:        client_publisher
# Purpose:     Publisher classes linked to TCP client
#
# Author:      Laurent Carré
#
# Created:     16/12/2021
# Copyright:   (c) Laurent Carré Sterwen Technolgy 2021
# Licence:     <your licence>
#-------------------------------------------------------------------------------
import socket
import logging

from .publisher import Publisher
from navigation_server.router_common import NavGenericMsg, N2K_MSG, NULL_MSG, NavThread
from .IPCoupler import TCPBufferedReader
from .coupler import Coupler, CouplerWriteError
from .nmea0183_msg import process_nmea0183_frame
from .nmea2000_msg import fromPGDY, fromPGNST, N2KEncodeError

_logger = logging.getLogger("ShipDataServer."+__name__)


class NMEAPublisher(Publisher):
    """
    Manages the publication of NMEA messages using a specified client.

    The NMEAPublisher class facilitates transmission of NMEA messages, handling
    initialization with specific couplers and filters, and ensuring proper cleanup
    of client resources. This class acts as a specialized implementation of the
    Publisher base class, enabling integration with client-defined behavior for
    NMEA message communication.

    Attributes
    ----------
    _client : ClientType
        The client object responsible for sending NMEA messages.

    Methods
    -------
    process_msg(msg: NavGenericMsg) -> bool
        Processes an incoming message, handling errors related to missing payloads
        and forwarding valid payloads to the client.

    last_action()
        Performs necessary cleanup operations, such as closing the client
        connection.

    descr() -> str
        Provides a description of the client associated with the publisher.
    """
    def __init__(self, client, couplers: list, filters):

        super().__init__(None, internal=True, couplers=couplers, name=client.descr(), filters=filters)
        self._client = client
        client.set_publisher(self)
        _logger.info("NMEA Publisher %s created" % self.object_name())

    def process_msg(self, msg: NavGenericMsg):
        if msg.raw is None:
            _logger.error("No transparent payload available for %s => wrong nmea2000 mode on input ?" % msg.dump())
            return True
        return not self._client.send(msg.raw)

    def last_action(self):
        if not self._stopflag:
            self._client.close()

    def descr(self):
        return self._client.descr()


class NMEA2000DYPublisher(NMEAPublisher):
    """
    Handles the publishing of NMEA 2000 messages dynamically.

    This class inherits from the NMEAPublisher class and is specifically designed
    to manage and publish NMEA 2000 messages using Digital Yacht (DY) format.

    Attributes:
        client: Manages communication with the data recipient, allowing transmission of
            processed NMEA data.
        couplers: Provides additional functionality or behavioral changes in the
            data processing pipeline.
        filters: Filters incoming data messages to determine which should be processed
            and published.

    """
    def __init__(self, client, couplers, filters):
        super().__init__(client, couplers, filters)

    def process_msg(self, msg: NavGenericMsg):
        if msg.type == N2K_MSG:
            try:
                data = msg.msg.asPDGY()
            except N2KEncodeError:
                _logger.error("Error on message %s => ignoring - check configuration" % msg.msg.format2())
                return True
            return not self._client.send(data)
        else:
            return super().process_msg(msg)


class NMEA2000STPublisher(NMEAPublisher):

    def __init__(self, client, couplers, filters):
        super().__init__(client, couplers, filters)

    def process_msg(self, msg: NavGenericMsg):
        if msg.type == N2K_MSG:
            data = msg.msg.asPGNST()
            return not self._client.send(data)
        else:
            return super().process_msg(msg)


class NMEASender(NavThread):
    """
    Manages the transmission of NMEA messages over a specified network connection.

    This class handles the sending of NMEA messages over a TCP connection. It processes input
    messages, manages their dispatch through a coupler, and optionally forwards them to a
    publishing system. The sender operates in a threaded manner, allowing asynchronous
    handling and termination of message transmission. Various modes of message processing
    can be configured based on the chosen NMEA2000 mode. It provides utility to track
    message counts, manage silent periods, and reset internal counters.

    Attributes:
        msg_processing (dict): A dictionary mapping message processing modes to their
            corresponding processing functions.
    """
    msg_processing = {'transparent': process_nmea0183_frame,
                      'dyfmt': fromPGDY, 'stfmt': fromPGNST}

    def __init__(self, connection: socket, address, coupler: Coupler, nmea2000_mode, buffer_size, timeout):
        super().__init__(name="Sender-" + "%s:%d" % address, daemon=True)
        self._connection = connection
        self._address = address
        self._coupler = coupler
        self._buffer_size = buffer_size
        self._timeout = timeout
        self._publisher = None
        self._stop_flag = False
        self._msgcount = 0
        self._silent_count = 0
        self._msg_processing = self.msg_processing[nmea2000_mode]

    def add_publisher(self, publisher):
        self._publisher = publisher

    def nrun(self) -> None:
        reader = TCPBufferedReader(self._connection, b'\r\n', self._address, self._msg_processing,
                                   self._buffer_size, self._timeout)
        while not self._stop_flag:
            msg = reader.read()
            # print(msg.printable())
            if msg.type == NULL_MSG:
                break
            self._msgcount += 1
            # check that we can write and stop otherwise 24/05/18
            try:
                self._coupler.send_msg_gen(msg)
            except CouplerWriteError as err:
                _logger.error("NMEASender write error %s" % err)
                break
            if self._publisher is not None:
                self._publisher.publish(msg)
        _logger.info("Stopping %s" % self.name)
        reader.stop()
        self._connection.close()

    def stop(self):
        _logger.info("NMEASender stop request")
        self._stop_flag = True
        if self._publisher is not None:
            self._publisher.stop()
        self._connection.close()

    def reset_period(self):
        self._msgcount = 0
        self._silent_count = 0

    def msgcount(self):
        return self._msgcount

    def add_silent_period(self):
        self._silent_count += 1

    def silent_count(self):
        return self._silent_count


class ClientConnection:
    """
    Manages an individual client connection within a server system, facilitating
    communication, message counting, and handling associated resources.

    The `ClientConnection` class wraps a socket connection and provides methods
    for sending and receiving messages, tracking message counts, managing silent
    periods, and interacting with publishers. The class also handles cleanup
    processes such as closing connections and deregistering components tied to
    the connection.

    Attributes:
        _socket: The socket object representing the client connection.
        _address: Tuple containing the remote client IP address and port number.
        _server: The server instance managing the client connections.
        _totalmsg: Tracks the total number of messages sent over this connection.
        _total_recmsg: Tracks the total number of messages received from this connection.
        _periodmsg: Tracks the number of messages sent or received during a specific period.
        _silent_count: Counts the number of consecutive silent periods of the connection.
        _publisher: The associated publisher instance managing outgoing data streams.
        _sender: The sender thread or mechanism used for handling outgoing messages.
    """
    def __init__(self, connection, address, server):
        self._socket = connection
        self._address = address
        self._server = server
        self._totalmsg = 0
        self._total_recmsg = 0
        self._periodmsg = 0
        self._silent_count = 0
        # a single publisher is enough (24/05/18)
        self._publisher = None
        self._sender = None

    def send(self, msg):
        try:
            self._socket.sendall(msg, socket.MSG_DONTWAIT)
            self._totalmsg += 1
            self._periodmsg += 1
            return False
        except OSError as e:
            _logger.warning(
                "Client:send Error writing data on %s:%d connection:%s => STOP" % (self._address[0], self._address[1], str(e)))
            return True

    def get(self):
        try:
            msg = self._socket.recv(512)
            self._total_recmsg += 1
            self._periodmsg += 1
        except OSError as e:
            _logger.warning(
                "Error reading data on %s:%d connection:%s => STOP" % (self._address[0], self._address[1], str(e)))
            return None
        if len(msg) == 0:
            _logger.warning(
                "Error reading data on %s:%d no data => STOP" % (self._address[0], self._address[1]))
            return None
        _logger.debug("Msg received:%s"% msg.decode().strip('\n\r'))
        # print("Msg received:%s"% msg.decode().strip('\n\r'))
        return msg

    def close(self):
        self._close()
        self._server.remove_client(self._address)

    def _close(self):
        if self._sender is not None:
            self._sender.stop()
            # self._server.remove_sender()

        self._publisher.deregister()
        self._publisher.stop()
        self._socket.close()

    def reset_period(self):
        self._periodmsg = 0

    def msgcount(self):
        return self._periodmsg

    def total_msg(self):
        return self._totalmsg

    def remote_ip(self):
        return self._address[0]

    def remote_port(self):
        return self._address[1]

    def add_silent_period(self):
        self._silent_count += 1

    def silent_count(self):
        return self._silent_count

    def clear_silent_count(self):
        self._silent_count = 0

    def set_publisher(self, pub):
        self._publisher = pub

    @property
    def publisher(self):
        return self._publisher

    def descr(self):
        return "Connection %s:%d" % self._address

    def connection(self):
        return self._socket

    def address(self):
        return self._address

    def read_status(self):
        out = {}
        out['object'] = 'connection'
        out['name'] = self.descr()
        out['total_msg_in'] = self._total_recmsg
        out['total_msg_out'] = self._totalmsg
        if self._sender is None:
            out['sender'] = None
        else:
            out['sender'] = self._sender.name
            out['sender_running'] = self._sender.is_alive()
        out['number_publisher'] = len(self._pubs)
        return out
